/* Interbench by Con Kolivas <kernel@kolivas.org> */
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sched.h>
#include <time.h>
#include <errno.h>
#include <semaphore.h>
#include <pthread.h>
#include <math.h>
#include <sys/utsname.h>

#define MAX_UNAME_LENGTH	100
#define MIN_BLK_SIZE	1024
#define MAX_LOG_LENGTH	255

struct user_data {
	unsigned long loops_per_ms;
	unsigned long ram;
	int duration;
	char logfilename[MAX_LOG_LENGTH];
	int log;
	char unamer[MAX_UNAME_LENGTH];
	char datestamp[13];
} ud = {
	0,
	0,
	30,
	"",
	1,
	"",
	""
};

struct sems {
	sem_t ready;
	sem_t start;
	sem_t stop;
	sem_t complete;
};

#define MAX_SAMPLES	10000	/* Limit the number of samples */
#define THREADS		9	/* The total number of different loads */

struct data_table {
	unsigned long total_latency;
	unsigned long max_latency;
	unsigned long nr_samples;
	unsigned long deadlines_met;
	unsigned long missed_latency;
	unsigned long missed_deadlines;
};

struct thread {
	void *(*name)(void *);
	char *label;
	int bench, load;
	pthread_t pthread;
	struct sems sem;
	struct data_table benchmarks[THREADS + 1], *dt;
	unsigned long sample[MAX_SAMPLES];
	unsigned long threadno;
};

void *emulate_none(void * t);
struct thread nonethread = {
	.label = "None", .name = emulate_none, .load = 1
};

void *emulate_audio(void * t);
struct thread audiothread = {
	.label = "Audio", .name = emulate_audio, .bench = 1, .load = 0
};

void *emulate_video(void * t);
struct thread videothread = {
	.label = "Video", .name = emulate_video, .bench = 1, .load = 0
};

void *emulate_x(void * t);
struct thread xthread = {
	.label = "X", .name = emulate_x, .bench = 1, .load = 1
};

void *emulate_game(void * t);
struct thread gamethread = {
	.label = "Gaming", .name = emulate_game, .bench = 1
};

void *emulate_burn(void * t);
struct thread burnthread = {
	.label = "Burn", .name = emulate_burn, .load = 1
};

void *emulate_write(void * t);
struct thread writethread = {
	.label = "Write", .name = emulate_write, .load = 1
};

void *emulate_read(void * t);
struct thread readthread = {
	.label = "Read", .name = emulate_read, .load = 1
};

void *emulate_ring(void * t);
struct thread ringthread = {
	.label = "Ring", .name = emulate_ring
};

struct thread *threadlist[THREADS] = {
	&nonethread,
	&audiothread,
	&videothread,
	&xthread,
	&gamethread,
	&burnthread,
	&writethread,	
	&readthread,
	&ringthread
};

void terminal_error(char *name)
{
	perror(name);
	exit (1);
}

void terminal_fileopen_error(FILE *fp, char *name)
{
	perror(name);
	if (fclose(fp) == -1)
		terminal_error("fclose");
	exit (1);
}

unsigned long long get_nsecs(struct timespec *myts)
{
	if (clock_gettime(CLOCK_MONOTONIC, myts))
		terminal_error("clock_gettime");
	return (myts->tv_sec * 1000000000 + myts->tv_nsec );
}

unsigned long get_usecs(struct timespec *myts)
{
	if (clock_gettime(CLOCK_MONOTONIC, myts))
		terminal_error("clock_gettime");
	return (myts->tv_sec * 1000000 + myts->tv_nsec / 1000 );
}

void burn_loops(unsigned long loops)
{
	unsigned long i;

	for (i = 0 ; i < loops ; i++);
}

/* Use this many usecs of cpu time */
void burn_usecs(unsigned long usecs)
{
	unsigned long ms_loops;

	ms_loops = ud.loops_per_ms / 1000 * usecs;
	burn_loops(ms_loops);
}

unsigned long periodic_schedule(struct thread *th, unsigned long run_usecs,
	unsigned long interval_usecs, unsigned long deadline)
{
	unsigned long current_time, latency;
	struct data_table *tb;
	struct timespec myts;

	latency = 0;
	tb = th->dt;

	current_time = get_usecs(&myts);
	/* First sample latency is ignored */
	if (!tb->nr_samples)
		current_time = deadline;
	if (current_time > deadline)
		latency = current_time - deadline;
	if (tb->nr_samples < MAX_SAMPLES)
		th->sample[tb->nr_samples] = latency;
	tb->total_latency += latency;
	latency += tb->missed_latency;
	if (latency > tb->max_latency)
		tb->max_latency = latency;

	burn_usecs(run_usecs);
	current_time = get_usecs(&myts);
	/*
	 * If we meet the deadline we move the deadline forward, otherwise
	 * we consider it a missed deadline and dropped frame etc.
	 */
	if (deadline + interval_usecs > current_time) {
		usleep(deadline + interval_usecs - current_time);
		tb->deadlines_met++;
		deadline += interval_usecs;
		tb->total_latency += tb->missed_latency;
		tb->missed_latency = 0;
	} else {
		/*
		 * Missed deadline is counted but only used to calculate total
		 * deadlines possible under no load.
		 */
		tb->missed_deadlines++;
		tb->missed_latency = interval_usecs;
		deadline += interval_usecs * 2;
	}
	return deadline;
}

void initialise_thread_data(struct thread *th)
{
	struct data_table *tb = th->dt;

	tb->max_latency =
	tb->total_latency =
	tb->deadlines_met =
	tb->missed_latency =
	tb->missed_deadlines =
	tb->nr_samples = 0;
}

void *emulate_none(void * t)
{
	struct thread *th = &nonethread;
	struct sems *s = &th->sem;

	t = 0;
	sem_post(&s->ready);
	sem_wait(&s->start);

	sem_wait(&s->stop);
	sem_post(&s->complete);
	return NULL;
}

#define audio_interval	(50000)
#define audio_run	(audio_interval / 20)
/* We emulate audio by using 5% cpu and waking every 50ms */
void *emulate_audio(void * t)
{
	unsigned long deadline;
	struct thread *th = &audiothread;
	struct sems *s = &th->sem;
	struct timespec myts;

	t = 0;
	sem_post(&s->ready);
	sem_wait(&s->start);
	deadline = get_usecs(&myts);

	while (1) {
		deadline = periodic_schedule(th, audio_run, audio_interval,
			deadline);
		if (!sem_trywait(&s->stop))
			goto out;
		th->dt->nr_samples++;
	}

out:
	sem_post(&s->complete);
	return NULL;
}

/* We emulate video by using 40% cpu and waking for 60fps */
#define video_interval	(1000000 / 60)
#define video_run	(video_interval * 40 / 100)
void *emulate_video(void * t)
{
	unsigned long deadline;
	struct thread *th = &videothread;
	struct sems *s = &th->sem;
	struct timespec myts;

	t = 0;
	sem_post(&s->ready);
	sem_wait(&s->start);
	deadline = get_usecs(&myts);

	while (1) {
		deadline = periodic_schedule(th, video_run, video_interval,
			deadline);
		if (!sem_trywait(&s->stop))
			goto out;
		th->dt->nr_samples++;
	}

out:
	sem_post(&s->complete);
	return NULL;
}

/*
 * We emulate X by running for a variable percentage of cpu from 0-100% 
 * in 1ms chunks.
 */
void *emulate_x(void * t)
{
	unsigned long deadline;
	struct thread *th = &xthread;
	struct sems *s = &th->sem;
	struct timespec myts;

	t = 0;
	sem_post(&s->ready);
	sem_wait(&s->start);
	deadline = get_usecs(&myts);

	while (1) {
		int i, j;
		for (i = 0 ; i <= 100 ; i++) {
			j = 100 - i;
			deadline = periodic_schedule(th, i * 1000, j * 1000,
				deadline);
			deadline += i * 1000;
			if (!sem_trywait(&s->stop))
				goto out;
			th->dt->nr_samples++;
		}
	}

out:
	sem_post(&s->complete);
	return NULL;
}

/* 
 * We emulate gaming by using 100% cpu and seeing how many frames (jobs
 * completed) we can do in that time. Deadlines are meaningless with 
 * unlocked frame rates
 */
#define game_interval	(10000)
#define game_run	(game_interval)
void *emulate_game(void * t)
{
	unsigned long deadline;
	struct thread *th = &gamethread;
	struct sems *s = &th->sem;
	struct timespec myts;

	t = 0;
	sem_post(&s->ready);
	sem_wait(&s->start);
	deadline = get_usecs(&myts);

	while (1) {
		deadline = periodic_schedule(th, game_run, game_interval,
			deadline);
		if (!sem_trywait(&s->stop))
			goto out;
		th->dt->nr_samples++;
	}

out:
	sem_post(&s->complete);
	return NULL;
}

/* Just burn cpu continuously */
void *emulate_burn(void * t)
{
	struct thread *th = &burnthread;
	struct sems *s = &th->sem;

	t = 0;
	sem_post(&s->ready);
	sem_wait(&s->start);
	while (1) {
		burn_loops(ud.loops_per_ms);
		if (!sem_trywait(&s->stop))
			goto out;
	}

out:
	sem_post(&s->complete);
	return NULL;
}

/* Write a file the size of ram continuously */
void *emulate_write(void * t)
{
	struct thread *th = &writethread;
	struct sems *s = &th->sem;
	FILE *fp;
	char *name = "interbench.write";
	void *buf = NULL;
	struct stat statbuf;
	unsigned long mem;

	t = 0;
	if (!(fp = fopen(name, "w")))
		terminal_error("fopen");
	if (stat(name, &statbuf) == -1)
		terminal_fileopen_error(fp, "stat");
	if (statbuf.st_blksize < MIN_BLK_SIZE)
		statbuf.st_blksize = MIN_BLK_SIZE;
	mem = ud.ram / (statbuf.st_blksize / 1024);	/* kilobytes to blocks */
	if (!(buf = calloc(1, statbuf.st_blksize)))
		terminal_fileopen_error(fp, "calloc");
	if (fclose(fp) == -1)
		terminal_error("fclose");
	sync();
	sem_post(&s->ready);
	sem_wait(&s->start);

	while (1) {
		unsigned int i;

		if (!(fp = fopen(name, "w")))
			terminal_error("fopen");
		if (stat(name, &statbuf) == -1)
			terminal_fileopen_error(fp, "stat");
		for (i = 0 ; i < mem; i++) {
			if (fwrite(buf, statbuf.st_blksize, 1, fp) != 1)
				terminal_fileopen_error(fp, "fwrite");
			if (!sem_trywait(&s->stop))
				goto out;
		}
		if (fclose(fp) == -1)
			terminal_error("fclose");
	}

out:
	if (fclose(fp) == -1)
		terminal_error("fclose");
	if (remove(name) == -1)
		terminal_error("remove");
	sync();
	sem_post(&s->complete);
	return NULL;
}

void create_read_file(void)
{
	unsigned int i;
	FILE *fp;
	char *name = "interbench.read";
	void *buf = NULL;
	struct stat statbuf;
	unsigned long mem;

	if (remove(name) == -1) {
		if (errno != ENOENT)
			terminal_error("remove");
	}
	if (!(fp = fopen(name, "w")))
		terminal_error("fopen");
	if (stat(name, &statbuf) == -1)
		terminal_fileopen_error(fp, "stat");
	if (statbuf.st_blksize < MIN_BLK_SIZE)
		statbuf.st_blksize = MIN_BLK_SIZE;
	if (!(buf = calloc(1, statbuf.st_blksize)))
		terminal_fileopen_error(fp, "calloc");
	mem = ud.ram / (statbuf.st_blksize / 1024);	/* kilobytes to blocks */

	for (i = 0 ; i < mem; i++) {
		if (fwrite(buf, statbuf.st_blksize, 1, fp) != 1)
			terminal_fileopen_error(fp, "fwrite");
	}
	if (fclose(fp) == -1)
		terminal_error("fclose");
	sync();
}

/* Read a file the size of ram continuously */
void *emulate_read(void * t)
{
	struct thread *th = &readthread;
	struct sems *s = &th->sem;
	char *name = "interbench.read";
	void *buf = NULL;
	struct stat statbuf;
	unsigned long mem, bsize;
	int tmp;

	t = 0;
	if ((tmp = open(name, O_RDONLY)) == -1) {
		if (errno == ENOENT) {
			create_read_file();
			if ((tmp = open(name, O_RDONLY)) == -1)
				terminal_error("open");
		} else
			terminal_error("open");
	}
	if (stat(name, &statbuf) == -1) 
		terminal_error("stat");
	if (statbuf.st_blksize < MIN_BLK_SIZE)
		statbuf.st_blksize = MIN_BLK_SIZE;
	bsize = statbuf.st_blksize;
	if (statbuf.st_size / 1024 / bsize != ud.ram / bsize)
		create_read_file();
	mem = ud.ram / (bsize / 1024);
	if (!(buf = malloc(bsize)))
		terminal_error("malloc");

	sem_post(&s->ready);
	sem_wait(&s->start);

	while (1) {
		int rd;

		/* 
		 * We have to read the whole file before quitting the load
		 * to prevent the data being cached for the next read. This
		 * is also the reason the file is the size of physical ram.
		 */
		while ((rd = read(tmp , buf, bsize)) > 0);
		if(!sem_trywait(&s->stop))
			goto out;
		if (lseek(tmp, (off_t)0, SEEK_SET) == -1)
			terminal_error("lseek");
	}

out:
	sem_post(&s->complete);
	return NULL;
}

void *emulate_ring(void * t)
{
	struct thread *th = &ringthread;
	struct sems *s = &th->sem;

	t = 0;
	sem_post(&s->ready);
	sem_wait(&s->start);
	while (1) {
		burn_loops(ud.loops_per_ms);
		if (!sem_trywait(&s->stop))
			goto out;
	}

out:
	sem_post(&s->complete);
	return NULL;
}

void calibrate_loop(void)
{
	unsigned long long start_time, loops_per_msec, run_time = 0;
	unsigned long loops;
	struct timespec myts;

	loops_per_msec = 100000;
redo:
	while (run_time / 1000 != 1000) {
		loops = loops_per_msec;
		start_time = get_nsecs(&myts);
		burn_loops(loops);
		run_time = get_nsecs(&myts) - start_time;
		loops_per_msec = (1000000 * loops_per_msec / run_time ? :
			loops_per_msec);
	}

	/* Rechecking after a pause increases reproducibility */
	sleep(1);
	loops = loops_per_msec;
	start_time = get_nsecs(&myts);
	burn_loops(loops);
	run_time = get_nsecs(&myts) - start_time;
	if (run_time / 1000 != 1000)
		goto redo;

	ud.loops_per_ms = loops_per_msec;
}

void start_thread(struct thread *th)
{
	sem_post(&th->sem.start);
}

void stop_thread(struct thread *th)
{
	sem_post(&th->sem.stop);
	sem_wait(&th->sem.complete);

	/* Kill the thread */
	if (pthread_join(th->pthread, NULL))
		terminal_error("pthread_join");
}

void init_sem(sem_t *sem)
{
	if (sem_init(sem, 1, 0))
		terminal_error("sem_init");
}

void initialise_thread(int i)
{
	struct thread *th = threadlist[i];

	/* Initialise the semaphores */
	init_sem(&th->sem.ready);
	init_sem(&th->sem.start);
	init_sem(&th->sem.stop);
	init_sem(&th->sem.complete);

	/* Create the threads */
	if (pthread_create(&th->pthread, NULL, th->name, (void*)0))
		terminal_error("pthread_create");

	//printf("pthread %lu\n",th->pthread);
	sem_wait(&th->sem.ready);
	/* 
	 * The data in benchmarks[THREADS] is unused, so it will be set to
	 * the actual load being run only if this thread is being benchmarked
	 */
	th->dt = &th->benchmarks[THREADS];
	initialise_thread_data(th);
	
}

void log_output(const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	if (vprintf(format, ap) == -1)
		terminal_error("vprintf");
	if (ud.log) {
		FILE *fp;

		if (!(fp = fopen(ud.logfilename, "a")))
			terminal_error("fopen");
		if (vfprintf(fp, format, ap) == -1)
			terminal_error("vpfrintf");
		if (fclose(fp) == -1)
			terminal_error("fclose");
	}
	va_end(ap);
}

void show_latencies(struct thread *th)
{
	struct data_table *tbi = &th->benchmarks[0],
		*tbj = th->dt;
	double average_latency, max_latency, deadlines_met, samples_met, sd;
	unsigned long nr_samples, noload_samples, noload_deadlines, s;
	long double variance = 0;

	/* 
	 * Total deadlines possible is calculated by deadlines met +
	 * deadlines missed under no load
	 */
	noload_deadlines = (tbi->deadlines_met + tbi->missed_deadlines) ? : 1;
	noload_samples = tbi->nr_samples;
	nr_samples = tbj->nr_samples;
	average_latency = tbj->total_latency / nr_samples;
	for (s = 0 ; s < nr_samples; s++) {
		long double diff = (long double)(average_latency -
			th->sample[s]) / 1000;
		variance += diff * diff;
	}
	variance /= (nr_samples - 1) ? : 1;
	sd = sqrtl(variance);

	average_latency /= 1000;

	/* 
	 * Landing on the boundary of a deadline can make loaded runs appear
	 * to do more work than unloaded due to tiny duration differences.
	 */
	if (nr_samples > noload_samples)
		nr_samples = noload_samples;
	samples_met = (double)nr_samples / (double)noload_samples * 100;
	if (samples_met > 100)
		samples_met = 100;
	max_latency = tbj->max_latency;
	max_latency /= 1000;
	deadlines_met = (double)tbj->deadlines_met /
		(double)noload_deadlines * 100;
	if (deadlines_met > 100)
		deadlines_met = 100;

	log_output("%7.1f +/- ", average_latency);
	log_output("%-9.1f", sd);

#if 0	/* max latency is often bogus because it is often one dropped frame wide */
	log_output("%7.23\t", max_latency);
#endif

	log_output("\t%4.3g", samples_met);
	log_output("\t%11.3g\n", deadlines_met);
}

void get_ram(void)
{
	FILE *meminfo;

	if(!(meminfo = fopen("/proc/meminfo", "r")))
		terminal_error("fopen");
	fscanf(meminfo, "MemTotal: %lu kB", &ud.ram);
	if (fclose(meminfo) == -1)
		terminal_error("fclose");
}

void get_fifo(void)
{
	struct sched_param sp;
	memset(&sp, 0, sizeof(sp));
	sp.sched_priority = 99;
	if (sched_setscheduler(0, SCHED_FIFO, &sp) == -1) {
		if (errno != EPERM)
			terminal_error("sched_setscheduler");
		fprintf(stderr, "could not set RT scheduling\n");
	}
}

void get_normal(void)
{
	struct sched_param sp;
	memset(&sp, 0, sizeof(sp));
	sp.sched_priority = 0;
	if (sched_setscheduler(0, SCHED_OTHER, &sp) == -1) {
		fprintf(stderr, "could not unset RT scheduling!\n");
	}
}

void bench(struct thread *th, int j)
{
	/* Point the data table to the appropriate load being tested */
	th->dt = &th->benchmarks[j];
	initialise_thread_data(th);
	get_fifo();
	start_thread(th);
	usleep(ud.duration * 1000000);
	stop_thread(th);
	get_normal();
	show_latencies(th);
}

void get_logfilename(void)
{
	struct tm *mytm;
	struct utsname buf;
	time_t t;
	int year, month, day, hours, minutes;

	time(&t);
	if (uname(&buf) == -1)
		terminal_error("uname");
	if (!(mytm = localtime(&t)))
		terminal_error("localtime");
	year = mytm->tm_year + 1900;
	month = mytm->tm_mon + 1;
	day = mytm->tm_mday;
	hours = mytm->tm_hour;
	minutes = mytm->tm_min;
	strncpy(ud.unamer, buf.release, MAX_UNAME_LENGTH);

	sprintf(ud.datestamp, "%2d%02d%02d%02d%02d",
		year, month, day, hours, minutes);
	snprintf(ud.logfilename, MAX_LOG_LENGTH, "%s.%s.log",
		ud.unamer, ud.datestamp);
}

void usage(void)
{
	fprintf(stderr, "interbench v0.15 by Con Kolivas\n");
	fprintf(stderr, "interbench [-l <int>] [-u] [-t <int] [-b] [-c] [-h]\n");
	fprintf(stderr, " -l\tUse <int> loops per sec (default: use saved benchmark)\n");
	fprintf(stderr, " -u\tImitate uniprocessor\n");
	fprintf(stderr, " -t\tSeconds to run each benchmark (default: 30)\n");
	fprintf(stderr, " -b\tBenchmark loops_per_ms even if it is already known\n");
	fprintf(stderr, " -c\tOutput to console only (default: use console and logfile)\n");
	fprintf(stderr, " -h\tShow this help\n");
}

int main(int argc, char **argv)
{
	int q, i, j, affinity, benchmark = 0;
	FILE *fp;
	/* 
	 *This file stores the loops_per_ms to be reused in a filename that
	 * can't be confused
	 */
	char *fname = "interbench.loops_per_ms";

	while ((q = getopt(argc, argv, "hl:ut:bc")) != -1) {
		switch (q) {
			case 'h':
				usage();
				return (0);
			case 'l':
				ud.loops_per_ms = atoi(optarg);
				break;
			case 'u':
				affinity = 1;
				break;
			case 't':
				ud.duration = atoi(optarg);
				break;
			case 'b':
				benchmark = 1;
				break;
			case 'c':
				ud.log = 0;
				break;
			default:
				usage();
				return (1);
		}
	}

	if (affinity) {
#ifdef CPU_SET	/* Current glibc expects cpu_set_t */
		cpu_set_t cpumask;

		CPU_ZERO(&cpumask);
		CPU_SET(0, &cpumask);
#else		/* Old glibc expects unsigned long */
		unsigned long cpumask = 1;
#endif
		if (sched_setaffinity(0, sizeof(cpumask), &cpumask) == -1) {
			if (errno != EPERM)
				terminal_error("sched_setaffinity");
			fprintf(stderr, "could not set cpu affinity\n");
		}
	}

	/* Make benchmark a multiple of 10 seconds for proper range of X loads */
	if (ud.duration % 10)
		ud.duration += 10 - ud.duration % 10;
	if (benchmark)
		ud.loops_per_ms = 0;
	/* 
	 * Try to get loops_per_ms from command line first, file second, and
	 * benchmark if not available.
	 */
	if (!ud.loops_per_ms) {
		if (benchmark)
			goto bench;
		if ((fp = fopen(fname, "r"))) {
			fscanf(fp, "%lu", &ud.loops_per_ms);
			if (fclose(fp) == -1)
				terminal_error("fclose");
			if (ud.loops_per_ms) {
				fprintf(stderr,
					"%lu loops_per_ms read from file interbench.loops_per_ms\n",
					ud.loops_per_ms);
				goto loops_known;
			}
		}
bench:
		fprintf(stderr, "loops_per_ms unknown; benchmarking...\n");

		/*
		 * To get as accurate a loop as possible we time it running
		 * SCHED_FIFO if we can
		 */
		get_fifo();
		calibrate_loop();
		get_normal();
	} else
		fprintf(stderr, "loops_per_ms specified from command line\n");

	if (!(fp = fopen(fname, "w"))) {
		fprintf(stderr, "Unable to write to file interbench.loops_per_ms\n");
		goto loops_known;
	}
	fprintf(fp, "%lu", ud.loops_per_ms);
	fprintf(stderr, "%lu loops_per_ms saved to file interbench.loops_per_ms\n",
		ud.loops_per_ms);
	if (fclose(fp) == -1)
		terminal_error("fclose");

loops_known:
	get_ram();
	get_logfilename();
	if (ud.log && !(fp = fopen(ud.logfilename, "w"))) {
		fprintf(stderr, "Unable to write to logfile\n");
		ud.log = 0;
	} else if (fclose(fp) == -1)
		terminal_error("fclose");
	log_output("Using %lu loops per ms, running every load for %d seconds\n",
		ud.loops_per_ms, ud.duration);
	log_output("Benchmarking kernel %s, at datestamp %s\n",
		ud.unamer, ud.datestamp);
	log_output("\n");

	for (i = 0 ; i < THREADS ; i++) {
		struct thread *thi = threadlist[i];

		threadlist[i]->threadno = i;
		if (!threadlist[i]->bench)
			continue;
		log_output("--- Benchmarking %s in the presence of loads ---\n", thi->label);
		log_output("\tLatency +/- SD (ms)  %% Desired CPU  %% Deadlines Met\n");
/*
		log_output("None\t");
		sleep(1);
		initialise_thread(i);
		bench(thi, i);
*/
		for (j = 0 ; j < THREADS ; j++) {
			struct thread *thj = threadlist[j];

			if (j == i || !threadlist[j]->load)
				continue;
			log_output("%s:\t", thj->label);
			sleep(1);
			initialise_thread(i);
			initialise_thread(j);
			start_thread(thj);
			bench(thi, j);
			stop_thread(thj);
		}
		log_output("\n");
	}

	return 0;
}
